unit SBWAV;
{****************************************************************************
** Unit to play 8bit uncompressed mono WAV files through the Sound Blaster **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}

interface
  function  SBFound : Boolean;
  procedure SBOpen;
  procedure SBClose;
  procedure SBPlay (FileName : String);
  procedure SBStop;
  function  SBLength : LongInt;
  function  SBPlaying : Boolean;

implementation

uses Dos, Crt;

{----------------------------------------------------------------------------
   INTERNAL STUFF
}

{A WAV header consits of several chunks:}
type
  RIFFChunkType = record
    RIFF : LongInt;
    NextChunkSize : LongInt;
    RIFFType : LongInt;
  end;
  fmtChunkType = record
    fmt : LongInt;
    fmtLength : LongInt;
    WaveType : Word;
    Channels : Word;
    SampleRate : LongInt;
    BytesPerSecond : LongInt;
    BlockAlignment : Word;
    BitResolution : Word;
  end;
  dataChunkType = record
    data : LongInt;
    dataLength : LongInt;
  end;

var
  {Sound Blaster settings}
  Base       : Word;    {Sound Blaster base address}
  DMA        : Byte;    {The DMA channel}
  IRQ        : Byte;    {The IRQ level}
  OldIRQ     : Pointer; {Pointer to old interrupt routines}

  {Memory buffer}
  RBuffer    : Word;    {Read buffer indicator}
  DMABuffer  : Pointer; {Pointer to DMA buffer}
  Page,
  Offset     : Word;

  {Header}
  dataChunk  : dataChunkType;
  fmtChunk   : fmtChunkType;
  RIFFChunk  : RIFFChunkType;

  {File access}
  WAVFile    : File;    {Handle of the WAV file}
  ToBeRead,             {Amount of samples to be read from file}
  ToBePlayed : LongInt; {Amount of samples to be played}

  {Global indicator}
  Playing    : Boolean;

{****************************************************************************
** Checks to see if a Sound Blaster exists at a given address, returns     **
** true if Sound Blaster found, false if not.                              **
****************************************************************************}
function ResetDSP (Test : Word) : Boolean;
begin
  {Reset the DSP}
  Port [Test + $6] := 1;
  Delay (10);
  Port [Test + $6] := 0;
  Delay (10);
  {Check if reset was succesfull}
  if (Port [Test + $E] and $80 = $80) and (Port [Test + $A] = $AA) then begin
    {DSP was found}
    ResetDSP := true;
    Base := Test;
  end else
    {No DSP was found}
    ResetDSP := false;
end;

{****************************************************************************
** Send a byte to the DSP (Digital Signal Processor) on the Sound Blaster  **
****************************************************************************}
procedure WriteDSP (Value : byte);
begin
  {Wait for the DSP to be ready to accept data}
  while Port [Base + $C] and $80 <> 0 do;
  {Send byte}
  Port [Base + $C] := value;
end;

{****************************************************************************
** The DMA controller is programmed with a block length of 32K - the       **
** entire buffer. The DSP is instructed to play blocks of 16K and then     **
** generate an interrupt (which allows the program to load the next part   **
** that should be played).                                                 **
****************************************************************************}
procedure AutoInitPlayback;
begin
  case DMA of
    0..3 : begin
      Port [$0A] := 4 or DMA;           {Mask DMA channel}
      Port [$0C] := 0;                  {Clear byte pointer}
      Port [$0B] := $58 or DMA;         {Set mode}
      Port [DMA shl 1] := Lo(Offset);   {Write the offset to the DMA controller}
      Port [DMA shl 1] := Hi(Offset);
    end;
  end;
  {
    The mode consists of the following:
    $58+x = binary 01 01 10 xx
                   |  |  |  |
                   |  |  |  +- DMA channel
                   |  |  +---- Read operation (the DSP reads from memory)
                   |  +------- Auto init mode
                   +---------- Block mode
  }
  case DMA of
    0 : Port [$87] := Page;            {Write the page to the DMA controller}
    1 : Port [$83] := Page;            {Write the page to the DMA controller}
    3 : Port [$82] := Page;            {Write the page to the DMA controller}
  end;

  case DMA of
    0..3 : begin
      Port [DMA shl 1 + 1] := $FF;     {Set the block length to $7FFF = 32 Kbyte}
      Port [DMA shl 1 + 1] := $7F;
      Port [$0A] := DMA;               {Unmask DMA channel}
    end;
  end;

  WriteDSP ($48);                      {Set the block length to $3FFF bytes = 16 Kbyte}
  WriteDSP ($FF);
  WriteDSP ($3F);

  WriteDSP ($1C);                      {DSP-command 1Ch - Start auto-init playback}
end;

procedure SingleCyclePlayback;
var
  BufOfs : Word;

begin
  dec (ToBePlayed);
  BufOfs := Offset + RBuffer shl 14;

  Port [$0A] := 4 or DMA;               {Mask DMA channel}
  Port [$0C] := 0;                      {Clear byte pointer}
  Port [$0B] := $48 or DMA;             {Set mode}
  Port [DMA shl 1] := Lo(BufOfs);       {Write the offset to the DMA controller}
  Port [DMA shl 1] := Hi(BufOfs);
  {
    The mode consists of the following:
    $48+x = binary 01 00 10 xx
                   |  |  |  |
                   |  |  |  +- DMA channel
                   |  |  +---- Read operation (the DSP reads from memory)
                   |  +------- Single cycle mode
                   +---------- Block mode
  }
  case DMA of
    0 : Port [$87] := Page;            {Write the page to the DMA controller}
    1 : Port [$83] := Page;            {Write the page to the DMA controller}
    3 : Port [$82] := Page;            {Write the page to the DMA controller}
  end;

  {Set the block length}
  Port [DMA shl 1 + 1] := lo (ToBePlayed);
  Port [DMA shl 1 + 1] := hi (ToBePlayed);
  Port [$0A] := DMA;                   {Unmask DMA channel}

  {DSP-command 14h - 8bit single cycle playback}
  WriteDSP ($14);
  WriteDSP (lo (ToBePlayed));
  WriteDSP (hi (ToBePlayed));

  {Nothing left to play}
  ToBePlayed := 0;
end;

{****************************************************************************
** Loads one half of the DMA buffer from the file                          **
****************************************************************************}
procedure ReadBuffer (Buffer : Word);
begin
  {If the remaining part of the file is smaller than 16K,
  load it and fill out with silence}
  if ToBeRead <= 0 then exit;
  if ToBeRead < 16384 then begin
    fillChar (Mem [Seg(DMABuffer^):Ofs(DMABuffer^) + Buffer shl 14], 16384, 128);
    BlockRead (WAVFile, Mem [Seg(DMABuffer^):Ofs(DMABuffer^) + Buffer shl 14], ToBeRead);
    ToBeRead := 0;
  end else begin
    BlockRead (WAVFile, Mem [Seg(DMABuffer^):Ofs(DMABuffer^) + Buffer shl 14], 16384);
    dec (ToBeRead, 16384);
  end;
end;

{****************************************************************************
** IRQ service routine - this is called when the DSP has finished playing  **
** a block                                                                 **
****************************************************************************}
procedure ServiceIRQ; interrupt;
var
  Temp : Byte;

begin
  {Relieve DSP, 8bit port}
  Temp := Port [Base + $E];
  {Acknowledge hardware interrupt}
  Port [$20] := $20;
  {Acknowledge cascade interrupt for IRQ 2, 10 and 11}
  if IRQ in [2, 10, 11] then Port [$A0] := $20;
  {Take appropriate action for buffers}
  if Playing then begin
    dec (ToBePlayed, 16384);
    if ToBePlayed > 0 then begin
      ReadBuffer (RBuffer);
      if ToBePlayed <= 16384 then begin
        RBuffer := RBuffer xor 1;
        SingleCyclePlayback;
      end else if ToBePlayed <= 32768 then begin
        WriteDSP ($DA);
      end;
    end else begin
      Playing := false;
    end;
  end;

  RBuffer := RBuffer xor 1;
end;

{****************************************************************************
** This procedure allocates 32K of memory to the DMA buffer and makes sure **
** that no page boundary is crossed                                        **
****************************************************************************}
procedure AssignBuffer;
var
  TempBuf       : Pointer; {Temporary pointer}
  LinearAddress : LongInt;
  Page1, Page2  : Word;

begin
  {Assign 32K of memory}
  GetMem (TempBuf, 32768);

  {Calculate linear address}
  LinearAddress := Seg (TempBuf^);
  LinearAddress := LinearAddress shl 4 + Ofs (TempBuf^);
  {Calculate page at start of buffer}
  Page1 := LinearAddress shr 16;
  {Calculate page at end of buffer}
  Page2 := (LinearAddress + 32767) shr 16;

  {Check to see if a page boundary is crossed}
  if (Page1 <> Page2) then begin
    {If so, assign another part of memory to the buffer}
    GetMem (DMABuffer, 32768);
    FreeMem (TempBuf, 32768);
  end else begin
    {otherwise, use the part we've already allocated}
    DMABuffer := TempBuf;
  end;

  {Convert pointer to linear address}
  LinearAddress := Seg (DMABuffer^);
  LinearAddress := LinearAddress shl 4 + Ofs (DMABuffer^);
  Page := LinearAddress shr 16;
  Offset := LinearAddress and $FFFF;
end;

{****************************************************************************
** This procedure checks the possible addresses to see whether a Sound     **
** Blaster is installed.                                                   **
****************************************************************************}
procedure FindSB;
var
  Temp : Byte;
  BLASTER : String;

begin
  {Nothing found yet}
  Base := 0;

  {Check for Sound Blaster, address: ports 210, 220, 230, 240, 250, 260 or 280}
  for Temp := 1 to 8 do begin
    if Temp <> 7 then
    if ResetDSP ($200 + Temp shl 4) then Break;
  end;
  if Temp = 9 then Exit;

  {Search for IRQ and DMA entry in BLASTER environment string}
  BLASTER := GetEnv ('BLASTER');
  DMA := 0;
  for Temp := 1 to length (BLASTER) do
    if UpCase (BLASTER [Temp]) = 'D' then
      DMA := Ord (BLASTER [Temp + 1]) - Ord ('0');
  for Temp := 1 to length (BLASTER) do
    if UpCase (BLASTER [Temp]) = 'I' then begin
      IRQ := Ord (BLASTER [Temp + 1]) - Ord ('0');
      if BLASTER [Temp + 2] <> ' ' then
        IRQ := IRQ * 10 + Ord (BLASTER [Temp + 2]) - Ord ('0');
    end;
end;

{****************************************************************************
** This procedure sets up the program according to the values in IRQ and   **
** DMA.                                                                    **
****************************************************************************}
procedure InitIRQandDMA;
begin
  {Save old IRQ vector}
  case IRQ of
    2 : GetIntVec($71, OldIRQ);
   10 : GetIntVec($72, OldIRQ);
   11 : GetIntVec($73, OldIRQ);
  else
    GetIntVec (8 + IRQ, OldIRQ);
  end;
  {Set new IRQ vector}
  case IRQ of
    2 : SetIntVec($71, Addr (ServiceIRQ));
   10 : SetIntVec($72, Addr (ServiceIRQ));
   11 : SetIntVec($73, Addr (ServiceIRQ));
  else
    SetIntVec (8 + IRQ, Addr (ServiceIRQ));
  end;
  {Enable IRQ}
  case IRQ of
    2 : Port [$A1] := Port [$A1] and not 2;
   10 : Port [$A1] := Port [$A1] and not 4;
   11 : Port [$A1] := Port [$A1] and not 8;
  else
    Port [$21] := Port [$21] and not (1 shl IRQ);
  end;
  if IRQ in [2, 10, 11] then Port [$21] := Port [$21] and not 4;
end;

{****************************************************************************
** This procedure releases the DMA channel and IRQ level                   **
****************************************************************************}
procedure ReleaseIRQandDMA;
begin
  {Free interrupt vectors used to service IRQs}
  case IRQ of
    2 : SetIntVec($71, OldIRQ);
   10 : SetIntVec($72, OldIRQ);
   11 : SetIntVec($73, OldIRQ);
  else
    SetIntVec (8 + IRQ, OldIRQ);
  end;

  {Mask IRQs}
  case IRQ of
    2 : Port [$A1] := Port [$A1] or 2;
   10 : Port [$A1] := Port [$A1] or 4;
   11 : Port [$A1] := Port [$A1] or 8;
  else
    Port [$21] := Port [$21] or (1 shl IRQ);
  end;
  if IRQ in [2, 10, 11] then Port [$21] := Port [$21] or 4;
end;

{----------------------------------------------------------------------------
   EXTERNALLY VISIBLE FUNCTIONS
}

{****************************************************************************
** This procedure returns true if the Sound Blaster is present.            **
****************************************************************************}
function SBFound : Boolean;
begin
  SBFound := Base <> 0;
end;

{****************************************************************************
** This procedure initialises the Sound Blaster                            **
****************************************************************************}
procedure SBOpen;
begin
  {Set up IRQ and DMA channels}
  InitIRQandDMA;

  {Assign memory to the DMA Buffer}
  AssignBuffer;

  {Set read buffer to first buffer}
  RBuffer := 0;
  Playing := false;
end;

{****************************************************************************
** This procedure gracefully shuts down the unit.                          **
****************************************************************************}
procedure SBClose;
begin
  {Stop any operation in progress}
  if Playing then SBStop;
  {Release the memory buffer}
  FreeMem (DMABuffer, 32768);
  {Release the IRQ and DMA channels}
  ReleaseIRQandDMA;
end;

{****************************************************************************
** Starts playing a WAV file.                                              **
****************************************************************************}
procedure SBPlay (FileName : String);
var
  Before : LongInt;

begin
  {Don't play if there's no buffer}
  if (DMABuffer = nil) then Exit;

  {Start playback in buffer 0 and clear the buffer}
  RBuffer := 0;
  fillChar (DMABuffer^, 32768, 128);

  {Open the file for output}
  Assign (WAVFile, FileName);

  {$I-}
  Reset (WAVFile, 1);
  if (IOResult <> 0) then Exit;

  {Read RIFF chunk}
  BlockRead (WAVFile, RIFFChunk, sizeof (RIFFChunk));
  if (RIFFChunk.RIFF <> $46464952) or (RIFFChunk.RIFFType <> $45564157) then begin
    Close (WAVFile);
    exit;
  end;

  {Read fmt chunk}
  repeat
    Before := FilePos (WAVFile);
    BlockRead (WAVFile, fmtChunk, sizeof (fmtChunk));
    Seek (WAVFile, Before + fmtChunk.fmtLength + 8);
  until (fmtChunk.fmt = $20746D66);

  {Set playback frequency}
  WriteDSP ($40);
  WriteDSP (256 - 1000000 div fmtChunk.SampleRate);

  {Read data chunk}
  repeat
    Before := FilePos (WAVFile);
    BlockRead (WAVFile, dataChunk, sizeof (dataChunk));
    if (dataChunk.data <> $61746164) then
      Seek (WAVFile, Before + dataChunk.dataLength + 8);
  until (dataChunk.data = $61746164);
  ToBePlayed := dataChunk.dataLength;
  ToBeRead := ToBePlayed;

  {DSP-command D1h - Enable speaker}
  WriteDSP ($D1);

  {Read first bit of data}
  ReadBuffer (0);
  ReadBuffer (1);

  if (ToBeRead > 0) then AutoInitPlayback
  else SingleCyclePlayback;
  {$I+}

  Playing := true;
end;

{****************************************************************************
** Stops playback                                                          **
****************************************************************************}
procedure SBStop;
begin
  {Stops DMA-transfer}
  WriteDSP ($D0);

  {Playback has completed}
  Playing := false;

  {Close the file}
  {$I-}
  Close (WAVFile);
  if IOResult = 0 then;
  {$I+}
end;

{****************************************************************************
** This procedure returns the length of the WAV file in whole seconds.     **
****************************************************************************}
function SBLength : LongInt;
begin
  SBLength := dataChunk.dataLength div fmtChunk.BytesPerSecond;
end;

{****************************************************************************
** This procedure returns true if the WAV is still playing.                **
****************************************************************************}
function SBPlaying : Boolean;
begin
  SBPlaying := Playing;
end;

begin
  FindSB;
end.